@@ -14,7 +14,7 @@ use dataplane::api::Request;
 use dataplane::fetch::FetchablePartitionResponse;
 use dataplane::record::RecordSet;
 use dataplane::Isolation;
-use dataplane::smartstream::SmartStreamExtraParams;
+use dataplane::smartmodule::SmartModuleExtraParams;
 
 use flate2::{
     Compression,
@@ -31,15 +31,18 @@ use super::SpuServerApiKey;
 pub const WASM_MODULE_API: i16 = 11;
 pub const WASM_MODULE_V2_API: i16 = 12;
 
-// version for aggregator smartstream
+// version for aggregator SmartModule
 pub const AGGREGATOR_API: i16 = 13;
 
 // version for gzipped WASM payloads
 pub const GZIP_WASM_API: i16 = 14;
 
-// version for SmartStream array map
+// version for SmartModule array map
 pub const ARRAY_MAP_WASM_API: i16 = 15;
 
+// version for persistent SmartModule
+pub const SMART_MODULE_API: i16 = 16;
+
 /// Fetch records continuously
 /// Output will be send back as stream
 #[derive(Decoder, Encoder, Default, Debug)]
@@ -52,10 +55,17 @@ where
     pub fetch_offset: i64,
     pub max_bytes: i32,
     pub isolation: Isolation,
+    /// no longer used, but keep to avoid breaking compatibility, this will not be honored
+    // TODO: remove in 0.10
     #[fluvio(min_version = 11)]
     pub wasm_module: Vec<u8>,
+    // TODO: remove in 0.10
     #[fluvio(min_version = 12)]
-    pub wasm_payload: Option<SmartStreamPayload>,
+    pub wasm_payload: Option<LegacySmartModulePayload>,
+    #[fluvio(min_version = 16)]
+    pub smartmodule: Option<SmartModuleInvocation>,
+    #[fluvio(min_version = 16)]
+    pub derivedstream: Option<DerivedStreamInvocation>,
     pub data: PhantomData<R>,
 }
 
@@ -64,24 +74,50 @@ where
     R: Debug + Decoder + Encoder,
 {
     const API_KEY: u16 = SpuServerApiKey::StreamFetch as u16;
-    const DEFAULT_API_VERSION: i16 = GZIP_WASM_API;
+    const DEFAULT_API_VERSION: i16 = SMART_MODULE_API;
     type Response = StreamFetchResponse<R>;
 }
 
-/// The request payload when using a Consumer SmartStream.
+/// The request payload when using a Consumer SmartModule.
 ///
-/// This includes the WASM content as well as the type of SmartStream being used.
-/// It also carries any data that is required for specific types of SmartStreams.
+/// This includes the WASM content as well as the type of SmartModule being used.
+/// It also carries any data that is required for specific types of SmartModules.
+/// TODO: remove in 0.10
 #[derive(Debug, Default, Clone, Encoder, Decoder)]
-pub struct SmartStreamPayload {
-    pub wasm: SmartStreamWasm,
-    pub kind: SmartStreamKind,
-    pub params: SmartStreamExtraParams,
+pub struct LegacySmartModulePayload {
+    pub wasm: SmartModuleWasmCompressed,
+    pub kind: SmartModuleKind,
+    pub params: SmartModuleExtraParams,
+}
+
+/// The request payload when using a Consumer SmartModule.
+///
+/// This includes the WASM module name as well as the invocation being used.
+/// It also carries any data that is required for specific invocations of SmartModules.
+#[derive(Debug, Default, Clone, Encoder, Decoder)]
+pub struct SmartModuleInvocation {
+    pub wasm: SmartModuleInvocationWasm,
+    pub kind: SmartModuleKind,
+    pub params: SmartModuleExtraParams,
 }
 
-/// Indicates the type of SmartStream as well as any special data required
 #[derive(Debug, Clone, Encoder, Decoder)]
-pub enum SmartStreamKind {
+pub enum SmartModuleInvocationWasm {
+    /// Name of SmartModule
+    Predefined(String),
+    /// Compressed WASM module payload using Gzip
+    AdHoc(Vec<u8>),
+}
+
+impl Default for SmartModuleInvocationWasm {
+    fn default() -> Self {
+        Self::AdHoc(Vec::new())
+    }
+}
+
+/// Indicates the type of SmartModule as well as any special data required
+#[derive(Debug, Clone, Encoder, Decoder)]
+pub enum SmartModuleKind {
     Filter,
     Map,
     #[fluvio(min_version = ARRAY_MAP_WASM_API)]
@@ -89,9 +125,18 @@ pub enum SmartStreamKind {
     Aggregate {
         accumulator: Vec<u8>,
     },
+    #[fluvio(min_version = ARRAY_MAP_WASM_API)]
+    FilterMap,
+    #[fluvio(min_version = SMART_MODULE_API)]
+    Join(String),
+    #[fluvio(min_version = SMART_MODULE_API)]
+    JoinStream {
+        topic: String,
+        derivedstream: String,
+    },
 }
 
-impl Default for SmartStreamKind {
+impl Default for SmartModuleKind {
     fn default() -> Self {
         Self::Filter
     }
@@ -102,9 +147,9 @@ impl Default for SmartStreamKind {
 /// In a fetch request, a WASM module may be given directly in the request
 /// as raw bytes.
 ///
-// TODO ... or, it may be named and selected from the WASM store.
+// TODO: remove in 0.10
 #[derive(Clone, Encoder, Decoder)]
-pub enum SmartStreamWasm {
+pub enum SmartModuleWasmCompressed {
     Raw(Vec<u8>),
     /// compressed WASM module payload using Gzip
     #[fluvio(min_version = 14)]
@@ -127,7 +172,7 @@ fn unzip(compressed: &[u8]) -> io::Result<Vec<u8>> {
     Ok(buffer)
 }
 
-impl SmartStreamWasm {
+impl SmartModuleWasmCompressed {
     /// returns the gzip-compressed WASM module bytes
     pub fn to_gzip(&mut self) -> io::Result<()> {
         if let Self::Raw(raw) = self {
@@ -153,13 +198,13 @@ impl SmartStreamWasm {
     }
 }
 
-impl Default for SmartStreamWasm {
+impl Default for SmartModuleWasmCompressed {
     fn default() -> Self {
         Self::Raw(Vec::new())
     }
 }
 
-impl Debug for SmartStreamWasm {
+impl Debug for SmartModuleWasmCompressed {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         match self {
             Self::Raw(bytes) => f
@@ -174,6 +219,13 @@ impl Debug for SmartStreamWasm {
     }
 }
 
+///
+#[derive(Debug, Default, Clone, Encoder, Decoder)]
+pub struct DerivedStreamInvocation {
+    pub stream: String,
+    pub params: SmartModuleExtraParams,
+}
+
 #[derive(Encoder, Decoder, Default, Debug)]
 pub struct StreamFetchResponse<R>
 where
@@ -226,28 +278,29 @@ mod tests {
     use super::*;
 
     #[test]
-    fn test_encode_smartstreamkind() {
+    fn test_encode_smartmodulekind() {
         let mut dest = Vec::new();
-        let value: SmartStreamKind = SmartStreamKind::Filter;
+        let value: SmartModuleKind = SmartModuleKind::Filter;
         value.encode(&mut dest, 0).expect("should encode");
         assert_eq!(dest.len(), 1);
         assert_eq!(dest[0], 0x00);
     }
 
     #[test]
-    fn test_decode_smartstreamkind() {
+    fn test_decode_smartmodulekind() {
         let bytes = vec![0x01];
-        let mut value: SmartStreamKind = Default::default();
+        let mut value: SmartModuleKind = Default::default();
         value
             .decode(&mut std::io::Cursor::new(bytes), 0)
             .expect("should decode");
-        assert!(matches!(value, SmartStreamKind::Map));
+        assert!(matches!(value, SmartModuleKind::Map));
     }
 
     #[test]
-    fn test_encode_smartstreamwasm() {
+    fn test_encode_smartmodulewasm() {
         let mut dest = Vec::new();
-        let value: SmartStreamWasm = SmartStreamWasm::Raw(vec![0xde, 0xad, 0xbe, 0xef]);
+        let value: SmartModuleWasmCompressed =
+            SmartModuleWasmCompressed::Raw(vec![0xde, 0xad, 0xbe, 0xef]);
         value.encode(&mut dest, 0).expect("should encode");
         println!("{:02x?}", &dest);
         assert_eq!(dest.len(), 9);
@@ -263,16 +316,16 @@ mod tests {
     }
 
     #[test]
-    fn test_decode_smartstreamwasm() {
+    fn test_decode_smartmodulewasm() {
         let bytes = vec![0x00, 0x00, 0x00, 0x00, 0x04, 0xde, 0xad, 0xbe, 0xef];
-        let mut value: SmartStreamWasm = Default::default();
+        let mut value: SmartModuleWasmCompressed = Default::default();
         value
             .decode(&mut std::io::Cursor::new(bytes), 0)
             .expect("should decode");
         let inner = match value {
-            SmartStreamWasm::Raw(inner) => inner,
+            SmartModuleWasmCompressed::Raw(inner) => inner,
             #[allow(unreachable_patterns)]
-            _ => panic!("should decode to SmartStreamWasm::Raw"),
+            _ => panic!("should decode to SmartModuleWasm::Raw"),
         };
         assert_eq!(inner.len(), 4);
         assert_eq!(inner[0], 0xde);
@@ -287,9 +340,9 @@ mod tests {
         let value = DefaultStreamFetchRequest {
             topic: "one".to_string(),
             partition: 3,
-            wasm_payload: Some(SmartStreamPayload {
-                kind: SmartStreamKind::Filter,
-                wasm: SmartStreamWasm::Raw(vec![0xde, 0xad, 0xbe, 0xef]),
+            wasm_payload: Some(LegacySmartModulePayload {
+                kind: SmartModuleKind::Filter,
+                wasm: SmartModuleWasmCompressed::Raw(vec![0xde, 0xad, 0xbe, 0xef]),
                 ..Default::default()
             }),
             ..Default::default()
@@ -314,30 +367,32 @@ mod tests {
         value.decode(&mut std::io::Cursor::new(bytes), 12).unwrap();
         assert_eq!(value.topic, "one");
         assert_eq!(value.partition, 3);
-        let smartstream = match value.wasm_payload {
+        let sm = match value.wasm_payload {
             Some(wasm) => wasm,
             _ => panic!("should have smartstreeam payload"),
         };
-        let wasm = match smartstream.wasm {
-            SmartStreamWasm::Raw(wasm) => wasm,
+        let wasm = match sm.wasm {
+            SmartModuleWasmCompressed::Raw(wasm) => wasm,
             #[allow(unreachable_patterns)]
-            _ => panic!("should be SmartStreamWasm::Raw"),
+            _ => panic!("should be SmartModuleWasm::Raw"),
         };
         assert_eq!(wasm, vec![0xde, 0xad, 0xbe, 0xef]);
-        assert!(matches!(smartstream.kind, SmartStreamKind::Filter));
+        assert!(matches!(sm.kind, SmartModuleKind::Filter));
     }
 
     #[test]
     fn test_zip_unzip_works() {
         const ORIG_LEN: usize = 1024;
-        let orig = SmartStreamWasm::Raw(vec![0x01; ORIG_LEN]);
+        let orig = SmartModuleWasmCompressed::Raw(vec![0x01; ORIG_LEN]);
         let mut compressed = orig.clone();
         compressed.to_gzip().unwrap();
-        assert!(matches!(&compressed, &SmartStreamWasm::Gzip(ref x) if x.len() < ORIG_LEN));
+        assert!(
+            matches!(&compressed, &SmartModuleWasmCompressed::Gzip(ref x) if x.len() < ORIG_LEN)
+        );
         let mut uncompressed = compressed.clone();
         uncompressed.to_raw().unwrap();
         assert!(
-            matches!((&uncompressed, &orig), (&SmartStreamWasm::Raw(ref x), &SmartStreamWasm::Raw(ref y)) if x == y )
+            matches!((&uncompressed, &orig), (&SmartModuleWasmCompressed::Raw(ref x), &SmartModuleWasmCompressed::Raw(ref y)) if x == y )
         );
         assert_eq!(orig.get_raw().unwrap(), compressed.get_raw().unwrap());
     }
